import argparse
import os
import time

import cv2
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.by import By

URL = "SOME_URL_HERE"


def create_driver():
    options = Options()
    options.add_argument("--headless")
    options.add_argument("--disable-gpu")
    options.add_argument("--start-maximized")
    service = Service(executable_path="./chromedriver")
    driver = webdriver.Chrome(service=service, options=options)
    return driver


def capture_page(url: str, image_filename: str, html_filename: str, target_xpath: str):
    driver = create_driver()
    driver.get(url)
    el = driver.find_element(By.XPATH, target_xpath)
    total_height = el.size["height"] + 6000
    driver.set_window_size(1920, total_height)
    el.screenshot(image_filename)
    out_html = el.get_attribute("innerHTML")
    with open(html_filename, "w") as f:
        f.write(out_html)
    driver.close()
    driver.quit()


def calculate_image_similarity(image1: str, image2: str):
    prev_img = cv2.imread(image1)
    cur_img = cv2.imread(image2)

    if cur_img.shape != prev_img.shape:
        # cropping both image to fit smaller image
        if prev_img.shape[0] > cur_img.shape[0]:
            crop_height = cur_img.shape[0]
        else:
            crop_height = prev_img.shape[0]

        if prev_img.shape[1] > cur_img.shape[1]:
            crop_width = cur_img.shape[1]
        else:
            crop_width = prev_img.shape[1]

        prev_img = prev_img[:crop_height, :crop_width, :]
        cur_img = cur_img[:crop_height, :crop_width, :]
        print(f"Cropping image to height: {crop_height}, width: {crop_width}")

    height = cur_img.shape[0]
    width = cur_img.shape[1]

    errorL2 = cv2.norm(prev_img, cur_img, cv2.NORM_L2)
    similarity = 1 - errorL2 / (height * width)
    return similarity


def calculate_html_similarity(html1: str, html2: str):
    from difflib import SequenceMatcher

    with open(html1, "r") as f:
        html1 = f.readlines()
    with open(html2, "r") as f:
        html2 = f.readlines()

    return SequenceMatcher(None, html1, html2).ratio()


def detect_change(
    url: str,
    check_interval: int,
    image_similarity_threshold: float,
    html_similarity_threshold: float,
    targer_xpath: str,
):
    # create dir to isolate different url
    url_dir = url.replace("https://", "").replace("/", "_").replace(".", "_")
    if not os.path.isdir(url_dir):
        os.mkdir(url_dir)

    capture_page(url, f"{url_dir}/prev.png", f"{url_dir}/prev.html", targer_xpath)
    time.sleep(check_interval)
    capture_page(url, f"{url_dir}/cur.png", f"{url_dir}/cur.html", targer_xpath)
    html_similarity = calculate_html_similarity(
        f"{url_dir}/prev.html", f"{url_dir}/cur.html"
    )
    image_similarity = calculate_image_similarity(
        f"{url_dir}/prev.png", f"{url_dir}/cur.png"
    )
    while (image_similarity >= image_similarity_threshold) and (
        html_similarity >= html_similarity_threshold
    ):
        # print prev similarity
        print(
            f"HTML threshold({html_similarity_threshold}) not exceeded:",
            html_similarity,
        )
        print(
            f"Image threshold({image_similarity_threshold}) not exceeded:",
            image_similarity,
        )

        # pause and check again
        time.sleep(check_interval)

        # replace cur to prev for next check
        os.remove(f"{url_dir}/prev.png")
        os.rename(f"{url_dir}/cur.png", f"{url_dir}/prev.png")
        os.remove(f"{url_dir}/prev.html")
        os.rename(f"{url_dir}/cur.html", f"{url_dir}/prev.html")

        # capture and check again
        capture_page(url, f"{url_dir}/cur.png", f"{url_dir}/cur.html", targer_xpath)

        # html similarity
        html_similarity = calculate_html_similarity(
            f"{url_dir}/prev.html", f"{url_dir}/cur.html"
        )

        # image similarity
        image_similarity = calculate_image_similarity(
            f"{url_dir}/prev.png", f"{url_dir}/cur.png"
        )

    if html_similarity < html_similarity_threshold:
        print(
            f"HTML threshold({html_similarity_threshold}) exceeded: {html_similarity}"
        )
    if image_similarity < image_similarity_threshold:
        print(
            f"Image threshold({image_similarity_threshold}) exceeded: {image_similarity}"
        )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Script to detect changes on a website"
    )
    parser.add_argument("-u", "--url", help="Website URL to monitor", required=True)
    parser.add_argument(
        "-i",
        "--interval",
        help="Interval to check URL, unit is seconds, default to 86400 seconds, which is 1 day",
        default=60 * 60 * 24,
    )
    parser.add_argument(
        "-ht",
        "--html-threshold",
        help="HTML similarity threshold for website, range from 0 to 1, the higher the number the more similar the website, default to 0.9",
        default=0.9,
    )
    parser.add_argument(
        "-it",
        "--image-threshold",
        help="Image similarity threshold for website, range from 0 to 1, the higher the number the more similar the website, default to 0.9",
        default=0.9,
    )
    parser.add_argument(
        "-t",
        "--target-xpath",
        help="Target xpath to check for within the provided website.",
        default="/html/body",
    )
    args = vars(parser.parse_args())

    URL = args["url"]
    if (not URL.startswith("https://")) and (URL.startswith("www.")):
        URL = "https://" + URL
    check_interval = int(args["interval"])
    html_similarity_threshold = float(args["html_threshold"])
    image_similarity_threshold = float(args["image_threshold"])
    targer_xpath = str(args["targer_xpath"])

    print(
        f'Monitoring website on "{URL}" every {check_interval}s, html_similarity_threshold: {html_similarity_threshold}, image_similarity_threshold: {image_similarity_threshold}'
    )
    detect_change(
        url=URL,
        check_interval=check_interval,
        html_similarity_threshold=html_similarity_threshold,
        image_similarity_threshold=image_similarity_threshold,
        targer_xpath=targer_xpath,
    )



import os
import re
from tkinter.ttk import Style
from turtle import color

import graphviz
from IPython.display import Image


class SQLAnalyzer:
    def __init__(self, sql_filepath, min_dot=1) -> None:
        self.min_dot = min_dot
        self.sql_filepath = sql_filepath
        if sql_filepath.endswith(".sql"):
            self.sql_scripts = [sql_filepath]
        else:
            if not sql_filepath.endswith(os.path.sep):
                sql_filepath = sql_filepath + os.path.sep
            self.sql_scripts = [
                sql_filepath + x for x in os.listdir(sql_filepath) if x.endswith(".sql")
            ]

    def remove_leading_comments(self, line):
        return re.sub("(--|#)+.*$", "", line)

    def remove_capsulated_comments(self, query):
        return re.sub(r"/\*.*\*/", "", query)

    def read_lines(self, sql_filepath):
        with open(sql_filepath, "r") as f:
            lines = f.readlines()
            lines = [self.remove_leading_comments(x) for x in lines]
            one_liner = " ".join([x.upper().strip() for x in lines])
            queries = [x.strip() for x in one_liner.split(";") if x.strip() != ""]
            queries = [" ".join(x.split()) for x in queries]
        return queries

    def extract_statement(self, query):
        if self.min_dot == 0:
            table_regex = r"`?([A-Z0-9\-_]*\.?[A-Z0-9\-_]*\.?[A-Z0-9\-_]+)`?"
        elif self.min_dot == 1:
            table_regex = r"`?([A-Z0-9\-_]*\.?[A-Z0-9\-_]+\.[A-Z0-9\-_]+)`?"
        elif self.min_dot == 2:
            table_regex = r"`?([A-Z0-9\-_]+\.[A-Z0-9\-_]+\.[A-Z0-9\-_]+)`?"
        else:
            raise Exception("min_dot range must be between 0 and 2")
        # Source
        ## FROM (search for word after FROM with 2 dots in its name, with or without backticks)
        from_tables = re.findall(f"(?!DELETE )FROM {table_regex}", query)

        ## JOIN
        join_tables = re.findall(f"JOIN {table_regex}", query)

        # Output
        ## CREATE CREATE [ OR REPLACE ] [ TEMP | TEMPORARY ] TABLE [ IF NOT EXISTS ]
        create_tables = re.findall(
            f"CREATE (OR REPLACE )?(TEMP |TEMPORARY )?TABLE (IF NOT EXISTS )?{table_regex}",
            query,
        )
        create_tables = [x[3] for x in create_tables]

        ## INSERT INTO
        insert_tables = re.findall(f"INSERT (INTO )?{table_regex}", query)
        insert_tables = [x[1] for x in insert_tables]

        # Others
        ## DELETE FROM
        delete_tables = re.findall(f"DELETE (FROM )?{table_regex}", query)
        delete_tables = [x[1] for x in delete_tables]

        # define output
        query_source_tables = from_tables + join_tables
        query_output_tables = create_tables + insert_tables
        quert_delete_tables = delete_tables
        return query_source_tables, query_output_tables, quert_delete_tables

    def extract_sql(self, sql_filepath, groups=False):
        group_source_tables = []
        group_output_tables = []
        group_delete_tables = []
        queries = self.read_lines(sql_filepath)
        for query in queries:
            query = self.remove_capsulated_comments(query)
            (
                query_source_tables,
                query_output_tables,
                quert_delete_tables,
            ) = self.extract_statement(query)
            group_source_tables.append(list(set(query_source_tables)))
            group_output_tables.append(list(set(query_output_tables)))
            group_delete_tables.append(list(set(quert_delete_tables)))
        source_tables = list(set([y for x in group_source_tables for y in x]))
        output_tables = list(set([y for x in group_output_tables for y in x]))
        delete_tables = list(set([y for x in group_delete_tables for y in x]))
        if groups:
            return group_source_tables, group_output_tables, group_delete_tables
        else:
            source_tables.sort()
            output_tables.sort()
            delete_tables.sort()
            return {
                "source_tables": source_tables,
                "output_tables": output_tables,
                "delete_tables": delete_tables,
            }

    def get_dependencies(self):
        output = {}
        for i, sql_script in enumerate(self.sql_scripts):
            # script_tail_name = os.path.split(sql_script)[-1].replace(".sql", "")
            output[sql_script] = self.extract_sql(sql_script)
        return output

    def plot_inline(self):
        dot = graphviz.Digraph(strict=True)
        dot.attr(rankdir="LR")
        for i, sql_script in enumerate(self.sql_scripts):
            script_tail_name = os.path.split(sql_script)[-1].replace(".sql", "")
            (
                group_source_tables,
                group_output_tables,
                group_delete_tables,
            ) = self.extract_sql(sql_script, groups=True)
            with dot.subgraph(name=f"cluster_{i}") as sg:
                sg.attr(label=f"<<u>{script_tail_name}.sql</u>>", color="darkblue")
                k = 1
                for j, (source_tables, output_tables, delete_tables) in enumerate(
                    zip(group_source_tables, group_output_tables, group_delete_tables)
                ):
                    if len(source_tables) + len(output_tables) > 0:
                        with sg.subgraph(name=f"cluster_{k}") as ssg:
                            ssg.attr(
                                label=f"subquery_{k}", color="black", style="dashed"
                            )
                            ssg.node(f"{script_tail_name}_step_{k}", shape="box")
                            # create edges
                            for source_table in source_tables:
                                ssg.edge(source_table, f"{script_tail_name}_step_{k}")
                            for output_table in output_tables:
                                ssg.edge(f"{script_tail_name}_step_{k}", output_table)
                            for delete_table in delete_tables:
                                ssg.node(delete_table, color="darkred")
                                ssg.edge(f"{script_tail_name}_step_{k}", delete_table)
                        k += 1

        dot.render("output", format="png")
        os.remove("output")

        return Image(filename="output.png")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['melpoi', 'melpoi.scraping.website-monitor', 'melpoi.sql.SQLAnalyzer']

package_data = \
{'': ['*']}

install_requires = \
['graphviz>=0.20.1,<0.21.0',
 'ipython>=8.6.0,<9.0.0',
 'opencv-python>=4.6.0.66,<5.0.0.0',
 'selenium>=4.6.0,<5.0.0']

setup_kwargs = {
    'name': 'melpoi',
    'version': '0.0.11',
    'description': '',
    'long_description': '# melpoi\nmelpoi is a repository with some of the scripts me myself personally find useful. They can be a simple ETL script or even a web scraping script. This is mainly for me to have a centralized place to look for scripts that I need. Feel free to use them if they can help you, you can also submit a pull request to add new things into this that you feel others might benefit from it.\n\n# Categories\n- [Scraping](https://github.com/la0bing/melpoi/tree/main/melpoi/scraping)\n- [SQL](https://github.com/la0bing/melpoi/tree/main/melpoi/sql)\n',
    'author': 'Melvin Low',
    'author_email': 'la0bing07148@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
